var dyno= require('./lib/dyn.js'),
    diff = require('deep-diff').diff,
    uuid= require('node-uuid').v4,
    _= require('underscore'),
    cclone= require('circularclone'),
    async= require('async'); 

var _parser= require('./lib/parser'), 
    _finder= require('./lib/finder'),
    _refiner= require('./lib/refiner'),
    _index= require('./lib/indexer'),
    _modify= require('./lib/capacity');

const _nu= function (v)
      {
         return typeof v!='undefined';
      }, 
      _isobject= function (v)
      {
          return typeof v=='object'&&!Array.isArray(v);
      },
      _isobjectarr= function (v)
      {
          return typeof v=='object'&&Array.isArray(v)&&v.length>0&&typeof v[0]=='object';
      },
      _collect= function (consume)
      {
          return function (cons) 
          { 
                var c;

                if (!(c=consume[cons.table]))
                  c= consume[cons.table]= { read: 0, write: 0 };

                c.read+= cons.read;
                c.write+= cons.write;
          };
      };

var dyngo= module.exports= function (opts,cb)
{
   var defaults= { hints: true }; 

   if (!cb)
   {
     cb= opts;
     opts= defaults;
   }

   opts= opts || defaults;
   opts= _.defaults(opts,defaults);

   var dyn= dyno(opts.dynamo,_.extend(opts.tx || {},{ txTable: opts.txTable })),
       finder= _finder(dyn),
       parser= _parser(dyn,opts),
       db= _.extend({ _dyn: dyn },opts.tx,{ txTable: opts.txTable }),
       _alias= function (table)
       {
          return (opts.tables || {} )[table] || table;
       };

   db.cleanup= function (obj)
   {
      var p= dyn.promise('clean');

      process.nextTick(function ()
      {
          p.trigger.clean(cclone(obj,function (key,value)
          {
             if (key.indexOf&&key.indexOf('_')==0&&key!='_id')
               return undefined; 
             else
               return value;
          }));
      });

      return p;
   };


   var configureTable= function (table)
       {
            table.ensuredIndexes= [];

            table.find= function (cond,projection,identity)
            {
                var p, modifiers= {}, table= this;

                modifiers.$consistent= !!table.$consistent;

                p= dyn.promise(['results','count','end'],null,'consumed');

                process.nextTick(function ()
                {
                   parser
                   .parse(table,modifiers,cond,projection,identity)
                   .parsed(function (query)
                   {
                       refiner= _refiner(dyn,query),
                       cursor= finder.find(query);
                       cursor.chain(refiner);
                       refiner.chain(p);
                   })
                   .error(p.trigger.error);
                });

                p.sort= function (o)
                {
                  modifiers.orderby= o; 
                  return p;
                };

                p.limit= function (n)
                {
                  modifiers.limit= n; 
                  return p;
                };

                p.window= function (n)
                {
                  modifiers.window= n; 
                  return p;
                };

                p.skip= function (n)
                {
                  modifiers.skip= n; 
                  return p;
                };

                p.noderef= function ()
                {
                  modifiers.noderef= true; 
                  return p;
                };

                var origCount= _.bind(p.count,p);

                p.count= function (fn)
                {
                  if (fn)
                     origCount(fn);
                  else 
                     modifiers.count= true; 

                  return p;
                };

                return p;
            };

            table.findOne= function ()
            {
                var p, args= arguments, table= this;

                p= dyn.promise('result','notfound','consumed');

                table.find.apply(table,args).limit(1).results(function (items)
                {
                     if (items.length==0)
                       p.trigger.notfound();
                     else
                       p.trigger.result(items[0]); 
                })
                .consumed(p.trigger.consumed)
                .error(p.trigger.error);

                return p;
            };

            table.consistent= function ()
            {
                return _.extend({ $consistent: true }, table);
            };

            table.save= function (_obj)
            {
                var objs= _obj ? (Array.isArray(_obj) ? _obj : [_obj]) : [],
                    consume= {},
                    p= dyn.promise(null,'updatedsinceread','consumed'), found= false;

                process.nextTick(function ()
                {

                    async.forEach(objs,
                    function (obj,done)
                    {
                        var gops= {},
                            ops= gops[table._dynamo.TableName]= [],
                            _hashrange= function (obj)
                            {
                                obj._id= obj._id || uuid();
                                obj._pos= obj._pos || 0;
                                obj._rev= (obj._rev || 0)+1;
                                obj._refs= [];
                            },
                            _index= function (obj)
                            {
                                 table.indexes.forEach(function (index)
                                 {
                                    var iops= index.update(obj,'put') || {};

                                    _.keys(iops).forEach(function (table)
                                    {
                                       var tops= gops[table]= gops[table] || []; 
                                       tops.push.apply(tops,_.collect(iops[table],function (op) { op.index= true; return op; }));
                                       tops.index= true;
                                    });
                                 });
                            },
                            _remove= function (item)
                            {
                                ops.push({ op: 'del', item: { _id: item._id, _pos: item._pos } });

                                table.indexes.forEach(function (index)
                                {
                                    var iops= index.update(item,'del') || {};

                                    _.keys(iops).forEach(function (table)
                                    {
                                       var tops= gops[table]= gops[table] || []; 
                                       tops.push.apply(tops,_.collect(iops[table],function (op) { op.index= true; return op; }));
                                    });
                                });
                            },
                            _save= function (obj)
                            {
                               var _keys= _.keys(obj),
                                   _omit= ['_old'],
                                   diffs= diff(_.omit(obj._old,['_old']) || {},_.omit(obj,['_old']));

                               if (!diffs
                                    ||(obj._old || {_rev: 0})._rev<obj._rev) return;

                               _hashrange(obj);

                               _keys.forEach(function (key)
                               {
                                    if (key.indexOf('___')==0)
                                    {
                                      if (!_isobjectarr(obj[key.substring(3)]))
                                      {
                                          _omit.push(key);
                                          return;
                                      }
                                    }
                                    else
                                    if (key.indexOf('__')==0)
                                    {
                                      if (!_isobject(obj[key.substring(2)]))
                                      {
                                          _omit.push(key);
                                          return;
                                      }
                                    }
                                    
                                    var type= typeof obj[key];

                                    if (type=='object'&&!_.contains(['_old','_refs'],key))
                                    {
                                       var desc= obj[key];

                                       if (desc==null)
                                         _omit.push(key);
                                       else
                                       if (desc instanceof Date)
                                       { /* let dyn convert */ }
                                       else
                                       if (_.keys(desc).length==0)
                                         _omit.push(key);
                                       else
                                       if (Array.isArray(desc))
                                       {
                                           if (desc.length)
                                           {
                                               if (typeof desc[0]=='object')
                                               {
                                                   var _id= obj['___'+key]= obj['___'+key] || uuid();

                                                   if (obj._old)
                                                   {
                                                       var old= obj._old[key];

                                                       if (old&&old.length>desc.length)
                                                         old.forEach(function (oitem,idx)
                                                         {
                                                            if (oitem._id==_id)
                                                            {
                                                                if (!_.findWhere(desc,{ _pos: oitem._pos }))
                                                                  _remove(oitem);
                                                            }
                                                            else
                                                            {
                                                                if (!_.findWhere(desc,{ _pos: idx }))
                                                                  _remove({ _id: _id, _pos: idx, _ref: oitem._id+'$:$'+oitem._pos });         
                                                            }
                                                         });
                                                   }

                                                   desc.forEach(function (val, pos)
                                                   {
                                                      if (val._id&&val._id!=_id)
                                                      {
                                                         _save(val);
                                                         _save({ _id: _id, _pos: pos, _ref: val._id+'$:$'+val._pos });
                                                         obj._refs.push(val._id);
                                                      }
                                                      else
                                                      {
                                                         val._id= _id;

                                                         if (!isNaN(val._pos)&&val._pos!=pos)
                                                         {
                                                           delete val['_old'];
                                                           delete val['_rev'];
                                                           _remove(val);
                                                         }

                                                         val._pos= pos;
                                                         _save(val);
                                                      }
                                                   });

                                                   _omit.push(key);
                                               }
                                           }
                                           else
                                           {
                                              var _id= obj['___'+key];

                                              if (_id&&obj._old[key].length)
                                                obj._old[key].forEach(_remove);

                                              _omit.push(key);
                                              _omit.push('___'+key);
                                           }
                                       }
                                       else
                                       {
                                           _save(desc);
                                           obj['__'+key]= desc._id+'$:$'+desc._pos;
                                           obj._refs.push(desc._id);
                                           _omit.push(key);
                                       }
                                    } 
                                    else
                                    if (type=='string'&&!obj[key])
                                      _omit.push(key);
                                    else
                                    if (type=='number'&&isNaN(obj[key]))
                                      _omit.push(key);
                               });

                               if (!obj._refs.length) 
                                 delete obj['_refs'];
                               else
                                 obj._refs= _.uniq(obj._refs);

                               _index(obj); // index after _ fields are set so they are indexable too

                               ops.unshift({ op: 'put', item: obj, omit: _omit }); // let the aggregate op came first of "contained" objects, so that the aggrgate version protects the rest
                            },
                            _mput= function (gops,done)
                            {
                               async.forEach(_.keys(gops),
                               function (_table,done)
                               {
                                  var tops= gops[_table];

                                  async.forEachSeries(tops, // forEachSeries: when deleting elements from array i need deletes of old item _pos done before new item _pos put
                                  function (op,done)
                                  {
                                     var tab= dyn.table(_table),
                                         obj= op.item;
                                       
                                     if (op.index)
                                       tab.hash('_hash',obj._hash)
                                          .range('_range',obj._range);
                                     else
                                       tab.hash('_id',obj._id)
                                          .range('_pos',obj._pos);

                                     if (op.op=='put')
                                         tab.put(_.omit(obj,op.omit),
                                          function ()
                                          {
                                             obj._old= cclone(_.omit(obj,'_old'));
                                             done();
                                          },
                                          { expected: obj._old&&_nu(obj._old._rev) ? { _rev: obj._old._rev } : undefined })
                                          .consumed(_collect(consume))
                                          .error(done);
                                     else
                                     if (op.op=='del')
                                         tab.delete(done)
                                            .error(done);
                                     else
                                       done(new Error('unknown update type:'+op.op));
                                  },
                                  done);
                               },
                               done);
                            };


                        _save(obj);

                        _.keys(gops).forEach(function (table)
                        {
                            if (gops[table].length==0)
                              delete gops[table];
                            else
                              found= true;
                        });

                        if (found)
                          _mput(gops,done);
                        else
                          process.nextTick(done);

                    },
                    function (err)
                    {
                          p.trigger.consumed(consume);

                          if (err)
                          {
                            if (err.code=='notfound')
                              p.trigger.updatedsinceread();
                            else
                              p.trigger.error(err);
                          }
                          else
                              p.trigger.success();
                    });

                });

                return p;
            };

            table.ensureIndex= function (fields)
            {
                  var p= dyn.promise();

                  process.nextTick(function ()
                  {
                      var index= _index(dyn,table,fields,opts);

                      if (index)
                        index.ensure(function (err)
                        {
                             if (err)
                               p.trigger.error(err);
                             else
                             {
                               table.indexes.push(index);
                               table.ensuredIndexes.push(fields);
                               p.trigger.success();
                             }
                        });
                      else
                        p.trigger.error(new Error('no known index type can index those fields'));
                  });

                  return p;
            };

            table.remove= function (filter)
            {
                var p= dyn.promise(null,null,'consumed'),
                    found= false,
                    foundErr= true,
                    consume= {},
                    _consumed= function (cons)
                    {
                       consume.read+= cons.read;
                       consume.write+= cons.write;
                    },
                    _error= function (err)
                    {
                       foundErr= true;
                       p.trigger.consumed(consume);
                       p.trigger.error(err);
                    },
                    _success= _.after(2,function ()
                    {
                       p.trigger.consumed(consume);
                       p.trigger.success();
                    }),
                    cursor= table.find(filter,table.indexes.length ? undefined : { _id: 1, _pos: 1 }),
                    _deleteItem= function (obj,done)
                    {
                          async.parallel([
                          function (done)
                          {
                              async.forEach(table.indexes,
                              function (index,done)
                              {
                                   index.remove(obj).success(done).error(done).consumed(_collect(consume));
                              },done);
                          },
                          function (done)
                          {
                              dyn.table(table._dynamo.TableName)
                                 .hash('_id',obj._id)
                                 .range('_pos',obj._pos)
                                 .delete(done)
                                 .consumed(_collect(consume))
                                 .error(done);
                          }],
                          done);
                    };

                cursor.results(function (items)
                {
                    found= true;

                    async.forEach(items,_deleteItem,
                    function (err)
                    {
                       if (err)
                         cursor.trigger.error(err); 
                       else
                       if (!items.next)
                         _success();
                    });
                })
                .consumed(_consumed)
                .error(_error)
                .end(_success);

                return p;
            };

            table.update= function (query,update)
            {
                var p= dyn.promise(null,null,'consumed'),
                    cursor= table.consistent().find(query),
                    found= false,
                    consume= {},
                    _consumed= function (cons)
                    {
                        _.keys(cons).forEach(function (table)
                        {
                            var c, tcons= cons[table];

                            if (!(c=consume[table]))
                              c= consume[table]= { read: 0, write: 0 };

                            c.read+= tcons.read;
                            c.write+= tcons.write;
                        }); 
                    },
                    _error= function (err)
                    {
                       p.trigger.consumed(consume);
                       p.trigger.error(err);
                    },
                    _success= function ()
                    {
                       p.trigger.consumed(consume);
                       p.trigger.success();
                    },
                    _updateItem= function (item,done)
                    {
                       if (update.$set)
                         table.save(_.extend(item,update.$set))
                              .success(done)
                              .consumed(_consumed)
                              .error(done); 
                       else
                       if (update.$unset)
                         table.save(_.omit(item,_.keys(update.$unset)))
                              .success(done)
                              .consumed(_consumed)
                              .error(done); 
                       else
                         done(new Error('unknown update type')); 
                    },
                    _updateItems= function (items)
                    {
                       found= true;

                       async.forEach(items,_updateItem,
                       function (err)
                       {
                          if (err)
                            cursor.trigger.error(err);
                       }); 
                    };


                cursor
                     .results(_updateItems)
                     .consumed(_consumed)
                     .error(_error)
                     .end(_success);

                return p;
            };

            table.modify= function (read,write)
            {
                return _modify(dyn,table._dynamo.TableName,read,write);
            };

            table.drop= function ()
            {
                var p= dyn.promise(),
                    _success= function ()
                    {
                      delete db[_alias(table._dynamo.TableName)];
                      p.trigger.success();
                    },
                    _check= function ()
                    {
                          dyn.describeTable(table._dynamo.TableName,
                          function (err,data)
                          {
                              if (err)
                              {
                                  if (err.code=='ResourceNotFoundException')
                                    _success();
                                  else
                                    p.trigger.error(err);
                              }
                              else
                                setTimeout(_check,5000);
                          });
                    };
        
                async.forEach(table.indexes,
                function (index,done)
                {
                    index.drop(done);
                },
                function (err)
                {
                    if (opts.hints) console.log('This may take a while...'.yellow);

                    dyn.deleteTable(table._dynamo.TableName,function (err)
                    {
                        if (err)
                        {
                           if (err.code=='ResourceNotFoundException')
                             _success();
                           else
                             p.trigger.error(err);
                        }
                        else
                           setTimeout(_check,5000);
                    });
                });

                return p;
            };

            return table;
       },
       configureTables= function (cb)
       {
            var configure= function (tables)
                {
                    async.forEach(Object.keys(tables),
                    function (table,done)
                    {
                          dyn.describeTable(table,function (err,data)
                          {
                              if (!err)
                              {
                                var hash= _.findWhere(data.Table.KeySchema,{ KeyType: 'HASH' }),
                                    range= _.findWhere(data.Table.KeySchema,{ KeyType: 'RANGE' });

                                if (hash&&hash.AttributeName&&hash.AttributeName=='_id'&&range&&range.AttributeName=='_pos')
                                  db[tables[table]]= configureTable({ _dynamo: data.Table, indexes: [] });
                              }

                              done(err);
                          });
                    },
                    function (err)
                    {
                       cb(err,err ? null : db);          
                    });
                };

             if (opts.tables)
               configure(opts.tables);
             else
               dyn.listTables(function (err,list)
               {
                   if (err)
                     cb(err);
                   else
                   {
                       var tables= {};
                       list.forEach(function (table) { tables[table]= table; });
                       configure(tables);
                   }
               });
       };

   db.createCollection= function (name)
   { 
      var p= dyn.promise(),
          _success= function ()
          {
              dyn.describeTable(name,function (err,data)
              {
                  if (!err)
                  {
                    db[name]= configureTable({ _dynamo: data.Table, indexes: [] });
                    p.trigger.success();
                  }
                  else
                    p.trigger.error(err);

              });
          };

        if (opts.hints) console.log('This may take a while...'.yellow);

        dyn.table(name)
           .hash('_id','S')
           .range('_pos','N')
           .create(function check()
           {
              dyn.table(name)
                 .hash('_id','xx')
                 .query(function ()
              {
                 _success();
              })
              .error(function (err)
              {
                 if (err.code=='ResourceNotFoundException')
                   setTimeout(check,5000);
                 else
                 if (err.code=='notfound')
                   _success();
                 else
                   p.trigger.error(err);
              });
           })
           .error(function (err)
           {
               if (err.code=='ResourceInUseException')
                 p.trigger.error(new Error('the collection exists'));
               else
                 p.trigger.error(err);
           });

      return p;
   };

   db.ensureTransactionTable= function (opts)
   { 
      opts= _.defaults(opts || {},{ name: 'dyngo-transaction-table' });
      var p= dyn.promise(),
          _success= function ()
          {
              dyn.describeTable(opts.name,function (err,data)
              {
                  if (!err)
                  {
                    db.txTable= { _dynamo: data.Table, indexes: [] };
                    p.trigger.success();
                  }
                  else
                    p.trigger.error(err);
              });
          };

        if (opts.hints) console.log('This may take a while...'.yellow);

        dyn.table(opts.name)
           .hash('_id','S')
           .range('_item','S')
           .create(function check()
           {
              dyn.table(opts.name)
                 .hash('_id','xx')
                 .query(function ()
              {
                 _success();
              })
              .error(function (err)
              {
                 if (err.code=='ResourceNotFoundException')
                   setTimeout(check,5000);
                 else
                 if (err.code=='notfound')
                   _success();
                 else
                   p.trigger.error(err);
              });
           })
           .error(function (err)
           {
               if (err.code=='ResourceInUseException')
                 _success();
               else
                 p.trigger.error(err);
           });

      return p;
   };

   db.transaction= function (txOpts)
   {
         var p= dyn.promise('transaction',null,'consumed');

         process.nextTick(function ()
         {
             if (!db.txTable)
             {
               p.trigger.error(new Error('no transaction table defined'));
               return;
             }

             var tab= dyn.table(db.txTable._dynamo.TableName),
                 init= function (tx)
                 {               
                     dyn.table(db.txTable._dynamo.TableName)
                        .hash('_id',tx._id)
                        .range('_item','_')
                        .put(tx,function ()
                        {
                           var dopts= _.extend({ tx: tx, txTable: db.txTable },opts,txOpts);

                           dyngo(dopts,
                           function (err,tx)
                           {
                              if (err)
                              {  
                                p.trigger.error(err);
                                return;
                              }

                              _.filter(_.keys(tx),function (key) { return !!tx[key].find; })
                              .forEach(function (tableName)
                              {
                                  db[tableName].ensuredIndexes.forEach(tx[tableName].ensureIndex);
                              });

                              dopts.tx.transaction= _.bind(db.transaction,db);

                              tx.commit= function ()
                              {
                                  var p= dyn.promise('committed','rolledback','consumed'),
                                      _commit= function (cb)
                                      {
                                          dyn.table(db.txTable._dynamo.TableName)
                                             .hash('_id',tx._id)
                                             .range('_item','_')
                                             .updateItem({ update: { state: { action: 'PUT', value: 'committed' } },
                                                         expected: { state: 'pending' } },
                                             function ()
                                             {
                                                tx.state= 'committed';
                                                cb();   
                                             })
                                             .consumed(p.trigger.consumed)
                                             .error(function (err)
                                             {
                                                 if (err.code=='notfound')
                                                   p.trigger.rolledback(true);
                                                 else
                                                   p.trigger.error(err);
                                             });
                                      },
                                      _complete= function (cb)
                                      {
                                          var finished= _.after(2,cb);

                                          dyn.table(db.txTable._dynamo.TableName)
                                             .hash('_id',tx._id)
                                             .range('_item','target::','BEGINS_WITH')
                                             .query(function (items)
                                              {
                                                 async.forEach(items,
                                                 function (item,done)
                                                 {
                                                     var _item= item._item.split('::'),
                                                         table= _item[1],
                                                         hash= { attr: _item[2], value: _item[3] },
                                                         range= { attr: _item[4], value: _item[4]=='_pos' ? +_item[5] : _item[5] };

                                                     if (item._txOp=='delete')
                                                         dyn.table(table)
                                                            .hash(hash.attr,hash.value)
                                                            .range(range.attr,range.value)
                                                            .delete(function () { done(); },{ expected: { _tx: tx._id } })
                                                            .consumed(p.trigger.consumed)
                                                            .error(p.trigger.error);
                                                     else
                                                         dyn.table(table)
                                                            .hash(hash.attr,hash.value)
                                                            .range(range.attr,range.value)
                                                            .updateItem({ update: { _txTransient: { action: 'DELETE' },
                                                                                    _txApplied: { action: 'DELETE' },
                                                                                    _txDeleted: { action: 'DELETE' },
                                                                                    _txLocked: { action: 'DELETE' },
                                                                                    _tx: { action: 'DELETE' } } },
                                                            function () { done(); })
                                                            .consumed(p.trigger.consumed)
                                                            .error(p.trigger.error);
                                                      
                                                 },
                                                 function (err)
                                                 {
                                                     if (err)
                                                       p.trigger.error(err); 
                                                     else
                                                     if (!items.next)
                                                       finished();
                                                 });
                                              },
                                              { attrs: ['_id','_item','_txOp'],
                                           consistent: true })
                                             .error(p.trigger.error)
                                             .consumed(p.trigger.consumed)
                                             .end(finished);
                                      },
                                      _clean= function (cb)
                                      {
                                          dyn.table(db.txTable._dynamo.TableName)
                                             .hash('_id',tx._id)
                                             .range('_item','_')
                                             .updateItem({ update: { state: { action: 'PUT', value: 'completed' } },
                                                         expected: { state: 'committed' } },
                                             function ()
                                             {
                                                tx.state= 'completed';
                                                cb();   
                                             })
                                             .consumed(p.trigger.consumed)
                                             .error(p.trigger.error);
                                      };

                                  if (tx.state=='pending')
                                    _commit(function ()
                                    {
                                          _complete(function ()
                                          {
                                               _clean(p.trigger.committed);
                                          });
                                    });
                                  else
                                  if (tx.state=='committed')
                                    _complete(function ()
                                    {
                                         _clean(p.trigger.committed);
                                    });
                                  else
                                    p.trigger.error(new Error("Invalid transaction state: "+tx.state));

                                  return p;
                              };

                              tx.rollback= function ()
                              {
                                  var p= dyn.promise('rolledback',null,'consumed'),
                                      _rollback= function (cb)
                                      {
                                          var finished= _.after(2,function ()
                                              {
                                                  dyn.table(db.txTable._dynamo.TableName)
                                                     .hash('_id',tx._id)
                                                     .range('_item','_')
                                                     .updateItem({ update: { state: { action: 'PUT', value: 'rolledback' } },
                                                                 expected: { state: 'pending' } },
                                                     function ()
                                                     {
                                                        tx.state= 'rolledback';
                                                        cb();   
                                                     })
                                                     .consumed(p.trigger.consumed)
                                                     .error(p.trigger.error);
                                              });

                                          dyn.table(db.txTable._dynamo.TableName)
                                             .hash('_id',tx._id)
                                             .range('_item','target::','BEGINS_WITH')
                                             .query(function (items)
                                              {
                                                 async.forEach(items,
                                                 function (item,done)
                                                 {
                                                     var _item= item._item.split('::'),
                                                         table= _item[1],
                                                         hash= { attr: _item[2], value: _item[3] },
                                                         range= { attr: _item[4], value: _item[4]=='_pos' ? +_item[5] : _item[5] },
                                                         clean= function ()
                                                         {
                                                               dyn.table(table)
                                                                  .hash(hash.attr,hash.value)
                                                                  .range(range.attr,range.value)
                                                                  .updateItem({ update: { _txTransient: { action: 'DELETE' },
                                                                                          _txApplied: { action: 'DELETE' },
                                                                                          _txDeleted: { action: 'DELETE' },
                                                                                          _txLocked: { action: 'DELETE' },
                                                                                          _tx: { action: 'DELETE' } } },
                                                                   function () { done(); })
                                                                  .consumed(p.trigger.consumed)
                                                                  .error(p.trigger.error);
                                                         };

                                                     if (item._txOp=='put')
                                                       dyn.table(table)
                                                          .hash(hash.attr,hash.value)
                                                          .range(range.attr,range.value)
                                                          .get(function (item)
                                                           {
                                                               if (item._txTransient)
                                                                 dyn.table(table)
                                                                    .hash(hash.attr,hash.value)
                                                                    .range(range.attr,range.value)
                                                                    .delete(function () { done(); })
                                                                    .consumed(p.trigger.consumed)
                                                                    .error(p.trigger.error); 
                                                               else
                                                                   dyn.table(db.txTable._dynamo.TableName)
                                                                      .hash('_id',tx._id)
                                                                      .range('_item',['copy',
                                                                                      table,
                                                                                      hash.attr,
                                                                                      hash.value,
                                                                                      range.attr,
                                                                                      range.value].join('::'))
                                                                      .get(function (copy)
                                                                       {
                                                                              copy= _.omit(copy,['_id',
                                                                                                 '_item',
                                                                                                 '_txLocked',
                                                                                                 '_txApplied',
                                                                                                 '_txDeleted',
                                                                                                 '_txTransient',
                                                                                                 '_tx']);

                                                                              copy[hash.attr]= hash.value;
                                                                              copy[range.attr]= range.value;

                                                                              dyn.table(table)
                                                                                 .hash(hash.attr,hash.value)
                                                                                 .range(range.attr,range.value)
                                                                                 .put(copy,function () { done(); })
                                                                                 .consumed(p.trigger.consumed)
                                                                                 .error(p.trigger.error); 
                                                                       })
                                                                       .consumed(p.trigger.consumed)
                                                                       .error(p.trigger.error); 
                                                           })
                                                           .consumed(p.trigger.consumed)
                                                           .error(p.trigger.error); 
                                                     else
                                                       clean();
                                                 },
                                                 function (err)
                                                 {
                                                     if (err)
                                                       p.trigger.error(err); 
                                                     else
                                                     if (!items.next)
                                                       finished();
                                                 });
                                              },
                                              { attrs: ['_id','_item','_txOp'],
                                           consistent: true })
                                             .error(p.trigger.error)
                                             .consumed(p.trigger.consumed)
                                             .end(finished);
                                      };

                                  if (tx.state=='pending')
                                    _rollback(p.trigger.rolledback);
                                  else
                                    p.trigger.error(new Error("Invalid transaction state: "+tx.state));

                                  return p;
                              };

                              p.trigger.transaction(tx);
                           });
                        })
                        .consumed(p.trigger.consumed)
                        .error(p.trigger.error);

                };

                if (typeof txOpts=='string')
                  tab.hash('_id',txOpts)
                     .range('_item','_')
                     .get(init,{ consistent: true })
                     .chain(p);
                else
                {
                     if (opts.tx)
                       p.trigger.error(new Error('cannot start a transaction within a transaction'));
                     else
                       init({ _id: uuid(), _item: '_', state: 'pending' });
                }
         });

         return p;
   };

   configureTables(cb);

};
